/*
 * Cdrom for Psemu Pro like Emulators
 *
 * By: linuzappz <linuzappz@hotmail.com>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/cdrom.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>
#include <gtk/gtk.h>
#include <time.h>

typedef char HWND;

#include "PSEmu Plugin Defs.h"

char CdromDev[256];
long ReadMode;

#define DEV_DEF		"/dev/cdrom"
#define NORMAL		0
#define THREADED	1
#define READ_MODES	2
char *ReadModeNames[2] = { 
    "Normal",
    "Threaded - Faster"
};

#define DATA_SIZE	(CD_FRAMESIZE_RAW-12)

#define itob(i)		((i)/10*16 + (i)%10)		/* u_char to BCD */
#define btoi(b)		((b)/16*10 + (b)%16)		/* BCD to u_char */

static inline int msf_to_lba(char m, char s, char f) {
    return (((m * CD_SECS) + s) * CD_FRAMES + f) - CD_MSF_OFFSET;
}

union {
    struct cdrom_msf msf;
    unsigned char buf[CD_FRAMESIZE_RAW];
} cr;

unsigned char *cdbuffer = cr.buf + 12; /* skip sync data */

int cdHandle = -1;
int _Thread = 0;
pthread_t thread = -1;
int readed;

void *CdrThread(void *arg);

#define CreateThread() { \
	if (_Thread == 0) { _Thread = 1; pthread_create(&thread, NULL, CdrThread, NULL); } }

#define ExitThread() { \
	{ _Thread = 0; pthread_exit(NULL); } }

char *LibName = "LinuzAppz Cdr Driver";

const unsigned char version = 1; // PSEmu 1.x library
const unsigned char revision = VERSION;
const unsigned char build = BUILD;

long CDRstop(void);

void LoadConf();
void SaveConf();

char *PSEgetLibName(void) {
    return LibName;
}

unsigned long PSEgetLibType(void) {
    return PSE_LT_CDR;
}

unsigned long PSEgetLibVersion(void) {
    return version<<16|revision<<8|build;
}

long CDRinit(void) {
    LoadConf();

    return 0;
}

long CDRshutdown(void) {
    return 0;
}

long CDRopen(void) {

    if (cdHandle > 0) return 0; /* it's already open */
    cdHandle = open(CdromDev, O_RDONLY);
    if (cdHandle != -1) { // if we can't open the cdrom we'll works as a null plugin
	ioctl(cdHandle, CDROM_LOCKDOOR, NULL);
	ioctl(cdHandle, CDROMSTART, NULL);
    }
    else printf ("CDR: Could not open %s, working as a null plugin\n", CdromDev);

    return 0;
}

long CDRclose(void) {
    if (cdHandle < 1) return 0;
    CDRstop();
    close(cdHandle);
    cdHandle = -1;
    if (thread != -1) { pthread_cancel(thread); thread = -1; }

    return 0;
}

// return Starting and Ending Track
// buffer:
//	byte 0 - start track
//	byte 1 - end track
long CDRgetTN(unsigned char *buffer) {
    struct cdrom_tochdr toc;

    if (cdHandle < 1) {
	buffer[0] = 1;
	buffer[1] = 1;
	return 0;
    }

    if (ioctl(cdHandle,CDROMREADTOCHDR,&toc) == -1) return -1;

    buffer[0] = toc.cdth_trk0; // start track
    buffer[1] = toc.cdth_trk1; // end track

    return 0;
}

// return Track Time
// buffer:
//	byte 0 - frame
//	byte 1 - second
//	byte 2 - minute
long CDRgetTD(unsigned char track, unsigned char *buffer) {
    struct cdrom_tocentry entry;

    if (cdHandle < 1) {
	memset(buffer+1,0,3);
	return 0;
    }

    if (track == 0) track = 0xaa; // total time
    entry.cdte_track = track;
    entry.cdte_format = CDROM_MSF;

    if (ioctl(cdHandle, CDROMREADTOCENTRY, &entry) == -1) return -1;

    buffer[0] = entry.cdte_addr.msf.frame; /* frame */
    buffer[1] = entry.cdte_addr.msf.second; /* second */
    buffer[2] = entry.cdte_addr.msf.minute; /* minute */

    return 0;
}

// read track
// time:
//	byte 0 - minute
//	byte 1 - second
//	byte 2 - frame
// uses bcd format
long CDRreadTrack(unsigned char *time) {
    if (cdHandle < 1) {
	memset(cr.buf, 0, DATA_SIZE);
	cdbuffer = cr.buf;
	return 0;
    }

    //printf ("CDRreadTrack %d:%d:%d\n", btoi(time[0]), btoi(time[1]), btoi(time[2]));

    cr.msf.cdmsf_min0 = btoi(time[0]);
    cr.msf.cdmsf_sec0 = btoi(time[1]);
    cr.msf.cdmsf_frame0 = btoi(time[2]);

    if (ReadMode == NORMAL) {
	if (ioctl(cdHandle, CDROMREADRAW, &cr) == -1) return -1;
    } else { // THREADED
	readed = -1;
	CreateThread();
    }

    return 0;
}

void *CdrThread(void *arg) {
    ioctl(cdHandle, CDROMREADRAW, &cr);
    readed=1;
    ExitThread();
}

// return readed track
unsigned char *CDRgetBuffer(void) {
    if (ReadMode == THREADED) {
    	while (readed == -1) {}
	pthread_join(thread, NULL);
    }

    return cdbuffer;
}

// plays cdda audio
// sector:
//	byte 0 - minute
//	byte 1 - second
//	byte 2 - frame
// does NOT uses bcd format
long CDRplay(unsigned char *sector) {
    struct cdrom_msf addr;
    unsigned char ptmp[4];

    if (cdHandle < 1) return 0;

    // 0 is the last track of every cdrom, so play up to there
    if (CDRgetTD(0, ptmp) == -1) return -1;
    addr.cdmsf_min0 = sector[0];
    addr.cdmsf_sec0 = sector[1];
    addr.cdmsf_frame0 = sector[2];
    addr.cdmsf_min1 = ptmp[2];
    addr.cdmsf_sec1 = ptmp[1];
    addr.cdmsf_frame1 = ptmp[0];

    if (ioctl(cdHandle, CDROMPLAYMSF, &addr) == -1) return -1;

    return 0;
}

// stops cdda audio
long CDRstop(void) {
    struct cdrom_subchnl sc;

    if (cdHandle < 1) return 0;

    sc.cdsc_format = CDROM_MSF;
    if (ioctl(cdHandle, CDROMSUBCHNL, &sc) == -1) return -1;

    switch (sc.cdsc_audiostatus)  {
    	case CDROM_AUDIO_PAUSED:
    	case CDROM_AUDIO_PLAY: ioctl(cdHandle, CDROMSTOP); break;
    }

    return 0;
}

struct CdrStat {
    unsigned long Type;
    unsigned long Status;
    unsigned char Time[3]; // current playing time
};

// reads cdr status
// type:
// 	0x00 - unknown
//	0x01 - data
//	0x02 - audio
//	0xff - no cdrom
// status:
//	0x00 - unknown
//	0x01 - error
//	0x04 - seek error
//	0x10 - shell open
//	0x20 - reading
//	0x40 - seeking
//	0x80 - playing
// time:
//	byte 0 - minute
//	byte 1 - second
//	byte 2 - frame
long CDRgetStatus(struct CdrStat *stat) {
    struct cdrom_subchnl sc;
    int ret;

    memset(stat, 0, sizeof(struct CdrStat));

    if (cdHandle < 1) return -1;

    sc.cdsc_format = CDROM_MSF;
    if (ioctl(cdHandle, CDROMSUBCHNL, &sc) != -1)
	memcpy(stat->Time, &sc.cdsc_absaddr.msf, 3);

    ret = ioctl(cdHandle, CDROM_DISC_STATUS);
    switch (ret) {
	case CDS_AUDIO: stat->Type = 0x02; break;
	case CDS_DATA_1: case CDS_DATA_2:
	case CDS_XA_2_1: case CDS_XA_2_2:
	    stat->Type = 0x01; break;
    }
    ret = ioctl(cdHandle, CDROM_DRIVE_STATUS);
    switch (ret) {
	case CDS_NO_DISC: stat->Type = 0xff; /*break;
	case CDS_TRAY_OPEN:*/ stat->Status|=0x10; break;
    }

    switch (sc.cdsc_audiostatus)  {
	case CDROM_AUDIO_PLAY: stat->Status|=0x80; break;
    }

    return 0;
}

GtkWidget *CreateCombo(GtkWidget *Box, GList *glist) {
    GtkWidget *Combo = gtk_combo_new();
    gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(Combo)->entry), FALSE);
    if (glist != NULL)
	gtk_combo_set_popdown_strings(GTK_COMBO(Combo), glist);
    gtk_box_pack_start(GTK_BOX(Box), Combo, FALSE, FALSE, 0);
    gtk_widget_show(Combo);	
    return Combo;
}

GtkWidget *ConfDlg;
GtkWidget *Edit;
GtkWidget *Combo;
GList *glist;

void OnOk(GtkMenuItem *menuitem, gpointer userdata) {
    char *tmp;
    int i;

    tmp = gtk_entry_get_text(GTK_ENTRY(Edit));
    strcpy(CdromDev, tmp);
    tmp = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(Combo)->entry));
    for (i=0; i<READ_MODES; i++) if (!strcmp(ReadModeNames[i], tmp)) ReadMode = i;
    gtk_widget_hide(ConfDlg);
    gtk_widget_destroy(ConfDlg);
    SaveConf();
    gtk_main_quit();
}

void OnCancel(GtkMenuItem *menuitem, gpointer userdata) {
    gtk_widget_hide(ConfDlg);
    gtk_widget_destroy(ConfDlg);
    gtk_main_quit();
}

long CDRconfigure(void) {
    GtkWidget *Label;
    GtkWidget *Ok,*Cancel;
    GtkWidget *Box,*Box1,*BBox;
    int i;

    LoadConf();

    ConfDlg = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_widget_set_usize(ConfDlg, 260, 130);
    gtk_window_set_title(GTK_WINDOW(ConfDlg), "CDR Config Dialog");
    gtk_window_set_position(GTK_WINDOW(ConfDlg), GTK_WIN_POS_CENTER);
    gtk_container_set_border_width(GTK_CONTAINER(ConfDlg), 10);

    Box = gtk_vbox_new(0, 10);
    gtk_container_add(GTK_CONTAINER(ConfDlg), Box);
    gtk_widget_show(Box);

    Box1 = gtk_hbox_new(0, 0);
    gtk_box_pack_start(GTK_BOX(Box), Box1, FALSE, FALSE, 0);
    gtk_widget_show(Box1);

    Label = gtk_label_new("Cdrom Dev:");
    gtk_box_pack_start(GTK_BOX(Box1), Label, FALSE, FALSE, 0);
    gtk_widget_show(Label);

    Edit = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(Edit), CdromDev);
    gtk_box_pack_start(GTK_BOX(Box1), Edit, FALSE, FALSE, 20);
    gtk_widget_show(Edit);

    Box1 = gtk_vbox_new(0, 0);
    gtk_box_pack_start(GTK_BOX(Box), Box1, FALSE, FALSE, 0);
    gtk_widget_show(Box1);

    Label = gtk_label_new("Select Read Mode:");
    gtk_box_pack_start(GTK_BOX(Box1), Label, FALSE, FALSE, 0);
    gtk_widget_show(Label);

    glist = NULL;
    for (i=0; i<READ_MODES; i++) glist = g_list_append(glist, ReadModeNames[i]);
    Combo = CreateCombo(Box1, glist);
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(Combo)->entry), ReadModeNames[ReadMode]);

    BBox = gtk_hbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box), BBox, FALSE, FALSE, 0);
    gtk_widget_show(BBox);

    Ok = gtk_button_new_with_label("Ok");
    gtk_signal_connect(GTK_OBJECT(Ok), "clicked", GTK_SIGNAL_FUNC(OnOk), NULL);
    gtk_container_add(GTK_CONTAINER(BBox), Ok);
    GTK_WIDGET_SET_FLAGS(Ok, GTK_CAN_DEFAULT);
    gtk_widget_show(Ok);

    Cancel = gtk_button_new_with_label("Cancel");
    gtk_signal_connect(GTK_OBJECT(Cancel), "clicked", GTK_SIGNAL_FUNC(OnCancel), NULL);
    gtk_container_add(GTK_CONTAINER(BBox), Cancel);
    GTK_WIDGET_SET_FLAGS(Cancel, GTK_CAN_DEFAULT);
    gtk_widget_show(Cancel);

    gtk_widget_show(ConfDlg);

    gtk_main();

    return 0;
}

GtkWidget *AboutDlg;

void OnAboutOk(GtkMenuItem *menuitem, gpointer userdata) {
    gtk_widget_hide(AboutDlg);
    gtk_widget_destroy(AboutDlg);
    gtk_main_quit();
}

void CDRabout(void) {
    GtkWidget *Label;
    GtkWidget *Ok;
    GtkWidget *Box,*BBox;
    char AboutText[255];

    sprintf (AboutText,"%s %d.%d\n",LibName, revision, build);

    AboutDlg = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_widget_set_usize(AboutDlg, 260, 80);
    gtk_window_set_title(GTK_WINDOW(AboutDlg), "CDR About Dialog");
    gtk_window_set_position(GTK_WINDOW(AboutDlg), GTK_WIN_POS_CENTER);
    gtk_container_set_border_width(GTK_CONTAINER(AboutDlg), 10);

    Box = gtk_vbox_new(0, 0);
    gtk_container_add(GTK_CONTAINER(AboutDlg), Box);
    gtk_widget_show(Box);

    Label = gtk_label_new(AboutText);
    gtk_box_pack_start(GTK_BOX(Box), Label, FALSE, FALSE, 0);
    gtk_widget_show(Label);

    BBox = gtk_hbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box), BBox, FALSE, FALSE, 0);
    gtk_widget_show(BBox);

    Ok = gtk_button_new_with_label("Ok");
    gtk_signal_connect(GTK_OBJECT(Ok), "clicked", GTK_SIGNAL_FUNC(OnAboutOk), NULL);
    gtk_container_add(GTK_CONTAINER(BBox), Ok);
    GTK_WIDGET_SET_FLAGS(Ok, GTK_CAN_DEFAULT);
    gtk_widget_show(Ok);

    gtk_widget_show(AboutDlg);

    gtk_main();
}

long CDRtest(void) {
    cdHandle = open(CdromDev, O_RDONLY);
    if (cdHandle == -1) return -1;
    close(cdHandle);
    cdHandle = -1;
    return 0;
}

void LoadConf() {
    FILE *f;
    char cfg[255];

    sprintf (cfg,"%s/cdr.cfg", getenv("HOME"));
    f = fopen(cfg,"r");
    if (f == NULL) {
	strcpy(CdromDev, DEV_DEF);
	ReadMode = THREADED;
	return;
    }
    fscanf(f, "CdromDev = %s\n", CdromDev);
    fscanf(f, "ReadMode = %ld\n", &ReadMode);
    fclose(f);
}

void SaveConf() {
    FILE *f;
    char cfg[255];

    sprintf (cfg,"%s/cdr.cfg",getenv("HOME"));
    f = fopen(cfg,"w");
    if (f == NULL) return;
    fprintf(f, "CdromDev = %s\n", CdromDev);
    fprintf(f, "ReadMode = %ld\n", ReadMode);
    fclose(f);
}
